-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.enableFFB = true

-- The direction of the real wheel's FFB
M.wheelFFBaxisdir = -1

-- The FFB force
M.wheelFFBForce = 400

-- The FFB steady force coefficient (values from 0 to 10)
M.wheelFFBICoef = 10

-- The FFB steady force limits (in Newton)
M.wheelFFBILimit = 500

-- FFB damping percent (values from 0 to 1)
-- for example for 10% set it to 0.1
M.wheelFFBDampingPercent = 0

-- FFB smoothing
M.wheelFFBSmoothing = 1000

M.hydros = {}
M.lastFFBforce = 0
local sf = newExponentialSmoothing(M.wheelFFBSmoothing)
local gsforce = 0

local FFBHydro = nil
local FFBHydroRefL = 0
local FFBID    = -1 -- >=0 are valid IDs
local wheeldispl = 0
local wheelKP = 1
local FFBIntegral = 0
local FFBIclimit = 1
local invFFBHydroRefL = 1

local function nop()
end

local function toInputSpace(h, state)
    if state > h.center then
        return (state - h.center) * h.invMultOut
    else
        return (state - h.center) * h.invMultIn
    end
end

local function updateGFX(dt)
    -- update the source command value
    for i,h in pairs(M.hydros) do
        h.cmd = math.min(math.max(electrics.values[h.inputSource] or 0, h.inputInLimit), h.inputOutLimit) * h.inputFactor
        if h.cmd ~= h.inputCenter or h.analogue == true then
            h._inrate = h.inRate
            h._outrate = h.outRate
        else
            -- set autocenter rate
            h._inrate = h.autoCenterRate
            h._outrate = h.autoCenterRate
        end

        if h.cmd >= h.inputCenter then
            h.cmd = h.cOut + h.cmd * h.multOut
        else
            h.cmd = h.cIn + h.cmd * h.multIn
        end

        h.hydroDirState = toInputSpace(h, h.state)
    end

    wheelKP = M.wheelFFBaxisdir * M.wheelFFBForce
    FFBIclimit = M.wheelFFBILimit / (M.wheelFFBICoef * M.wheelFFBForce)
end

local function update(dts)
    -- state: the state of the hydro from -1 to 1
    -- cmd the input value
    -- note: state is scaled to the ratio as the last step
    local hcmd = -1
    for i,h in pairs(M.hydros) do
        -- slowly approach the desired value
        if h.cmd < h.state then
            h.state = math.max(h.state - dts * h._inrate, h.cmd)
        else
            h.state = math.min(h.state + dts * h._outrate, h.cmd)
        end

        --if inputSource ~= "steering" then
        --    print("state: "..h.state .. " - inlimit: " .. h.inLimit.. " - state: "..h.state .. " cmd:"..h.cmd .. " val = "..val)
        --end
        obj:setBeamLengthRefRatio(h.beam.cid, h.state)
    end

    if FFBID >= 0 then
        local prevdispl
        if obj:beamIsBroken(FFBHydro.beam.cid) then
            M.lastFFBforce = 0
        else
            prevdispl = wheeldispl
            wheeldispl = toInputSpace(FFBHydro, FFBHydro.state) - sf:get(toInputSpace(FFBHydro, obj:getBeamLength(FFBHydro.beam.cid) * invFFBHydroRefL))
            FFBIntegral = sign(FFBIntegral) * math.min(math.abs(FFBIntegral), FFBIclimit) + wheeldispl * dts
            M.lastFFBforce = wheelKP * (wheeldispl + M.wheelFFBICoef * FFBIntegral + M.wheelFFBDampingPercent * (wheeldispl - prevdispl))
        end

        obj:sendForceFeedback(FFBID, M.lastFFBforce)
    end
end

-- nop'ed functions
M.updateGFX = updateGFX
M.update = update

local function init()
    if v.data.hydros == nil then
        M.updateGFX = nop
        M.update = nop
        M.hydros = {}
        return
    end

    M.hydros = shallowcopy(v.data.hydros)

    local hydroCount = 0
    for i, h in pairs (M.hydros) do
        hydroCount = hydroCount + 1
        h.inputCenter = h.inputCenter * h.inputFactor
        h.inputInLimit = h.inputInLimit * h.inputFactor
        h.inputOutLimit = h.inputOutLimit * h.inputFactor
        local inputFactorSign = sign(h.inputFactor)

        if h.inputFactor < 0 then
            h.inputInLimit, h.inputOutLimit = h.inputOutLimit, h.inputInLimit
        end

        local inputMiddle = (h.inputOutLimit + h.inputInLimit) * 0.5
        if h.inputCenter >= inputMiddle then
            h.center = 1 + (h.outLimit - 1) * (h.inputCenter - inputMiddle) / (h.inputOutLimit - inputMiddle)
        else
            h.center = 1 - (1 - h.inLimit) * (inputMiddle - h.inputCenter) / (inputMiddle - h.inputInLimit)
        end
        h.state = h.center
        h.multOut = (h.outLimit - h.center) / (h.inputOutLimit - h.inputCenter)
        h.cOut = h.center - h.inputCenter * h.multOut
        h.multIn = (h.center - h.inLimit) / (h.inputCenter - h.inputInLimit)
        h.cIn = h.center - h.inputCenter * h.multIn
        h.cmd = h.inputCenter
        h.invMultOut = 1 / (h.outLimit - h.center) * inputFactorSign
        h.invMultIn = 1 / (h.center - h.inLimit) * inputFactorSign
        h._inrate = h.inRate
        h._outrate = h.outRate
        h.hydroDirState = 0
        
        -- little workaround
        if h.inputSource == "steering" then
            h.inputSource = "steering_input"
        end

        if h.inputSource == "steering_input" then
            h.steering = true
        end
    end

    if hydroCount == 0 then
        M.updateGFX = nop
        M.update = nop 
    end

    for i,h in pairs(M.hydros) do
        if h.steering then
            FFBHydro = h
            break
        end
    end
    FFBHydroRefL = obj:getBeamRefLength(FFBHydro.beam.cid)
    invFFBHydroRefL = 1 / FFBHydroRefL
    if M.wheelFFBDampingPercent < 0 then
        M.wheelFFBDampingPercent = 2 * 0.5 * math.sqrt(M.wheelFFBForce) / M.wheelFFBForce -- 0.5 is wheel's mass
    end
    sf:set(0)
    FFBIntegral = 0

    if M.enableFFB then
        FFBID = obj:setupForceFeedback("xaxis")
    end

    if FFBID >= 0 then
        logInfo("yay, force feedback is available :D")
    end
end

local function reset()
    for k,h in pairs(M.hydros) do
        h.state = h.center
        h.cmd = h.inputCenter
        h._inrate = h.inRate
        h._outrate = h.outRate
    end

    sf:set(0)
    FFBIntegral = 0
end

local function sendHydroStateToGUI()
    obj:executeJS("HookManager.trigger('HydrosUpdate', "..encodeJson(M.hydros)..");",0)
end

-- public interface
M.init = init
M.reset = reset
M.sendHydroStateToGUI = sendHydroStateToGUI

return M